package com.docmosis.example;


import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import com.docmosis.SystemManager;
import com.docmosis.document.DocumentProcessor;
import com.docmosis.document.ProcessingException;
import com.docmosis.document.converter.ConversionException;
import com.docmosis.document.converter.ConversionFormat;
import com.docmosis.document.converter.ConversionInstruction;
import com.docmosis.document.converter.NoConvertersRunningException;
import com.docmosis.document.converter.pool.ConverterPoolException;
import com.docmosis.template.TemplateStoreException;
import com.docmosis.template.population.Cleanable;
import com.docmosis.template.population.DataProvider;
import com.docmosis.template.population.DataProviderBuilder;
import com.docmosis.template.population.render.FieldDetails;
import com.docmosis.template.population.render.FieldRenderer;
import com.docmosis.template.population.render.RenderedField;
import com.docmosis.template.store.DropStoreHelper;
import com.docmosis.template.store.TemplateContext;
import com.docmosis.template.store.TemplateIdentifier;
import com.docmosis.template.store.TemplateStoreFactory;
import com.docmosis.util.FileUtilities;
import com.docmosis.util.logging.LogManager;
import com.docmosis.util.logging.Logger;

/**
 * Example code demonstrating some of the ways that Docmosis can be used.
 */
public class ExampleRenders
{
	private static final Logger log = LogManager.getLogger(ExampleRenders.class);

	private static final File STATS_FILE = new File("./stats/stats.csv");
	private static BufferedWriter statsWriter;
	
	private static void prepareTemplates()
	throws IOException, ConversionException, ConverterPoolException, TemplateStoreException, NoConvertersRunningException
	{
		DropStoreHelper storeHelper = new DropStoreHelper(TemplateStoreFactory.getStore());
		storeHelper.process(new File("."), false);
		log.info("Template(s) loaded into store");
	}
	
	private static void runExampleRender(DataProvider dp) 
	  throws IOException, ConversionException, ConverterPoolException, TemplateStoreException, ProcessingException
	{
		log.info("Performing a test render");
		TemplateIdentifier templateId = new TemplateIdentifier("exampleTemplate.doc", 
				new TemplateContext(""));
		
		ConversionInstruction instruction = new ConversionInstruction();
		instruction.setConversionFormats(new ConversionFormat[]{
				ConversionFormat.FORMAT_ODT,
				ConversionFormat.FORMAT_WORD,
				ConversionFormat.FORMAT_PDF,
		});
		instruction.setOutputFileName("exampleTemplate");
		FileOutputStream streamTo = new FileOutputStream("newDocuments.zip");
		try {
			long start = System.currentTimeMillis();
			DocumentProcessor.renderDoc(templateId, dp, instruction, streamTo);
			log.info("newDocuments.zip generated in " + (System.currentTimeMillis() - start) + " milliseconds (including wait time)");
		} finally {
			FileUtilities.close(streamTo);
		}
	}
	

	// Create the file for recording statistics.
	private static void createStatsFile() throws IOException
	{
		// Ensure directory structure.
		String dirName = "./stats";
		final File root = new File(dirName);
		if (!root.exists()) {
			if (!root.mkdir()) {
				throw new IOException("Unable to create dir:" + dirName);
			}
		}
		
		// Refresh file
		STATS_FILE.delete();
		STATS_FILE.createNewFile();
		statsWriter = new BufferedWriter(new FileWriter(STATS_FILE));
	}
	
	// Record the statistics to a file.
	private static void recordStatistics(long timeTaken) throws IOException
	{
		statsWriter.write(String.valueOf(timeTaken));
		statsWriter.write(',');
	}
	
	
	private static void renderDoc(File outputTo, boolean outputStats, DataProvider dp, ConversionInstruction instruction) 
		throws IOException, ProcessingException, TemplateStoreException
	{
		TemplateIdentifier templateId = 
			new TemplateIdentifier("exampleTemplate.doc", 
			new TemplateContext(""));
		
		FileOutputStream streamTo = new FileOutputStream(outputTo);
		try {
			long start = System.currentTimeMillis();
			DocumentProcessor.renderDoc(templateId, dp, instruction, streamTo);
			long timeTaken = (System.currentTimeMillis() - start);
			log.info("newDocuments.zip generated in " + timeTaken + " milliseconds (including wait time)");
			if (outputStats) {
				recordStatistics(timeTaken);
			}
		} finally {
			FileUtilities.close(streamTo);
		}
		
	}
	
	private static void runConcurrentRenders(final int nThreads,
											 final int nRendersPerThread,
											 final boolean outputStats,
											 final DataProvider dp) 
	throws IOException
	{
		try {
	    	long start = System.currentTimeMillis();
	    	final Thread[] threads = new Thread[nThreads];
	    	
			String dirName = "./renders";
			final File root = new File(dirName);
			if (!root.exists()) {
				if (!root.mkdir()) {
					throw new IOException("Unable to create dir:" + dirName);
				}
			}
			
			if (outputStats) {
				createStatsFile();
			}
			
			DirectoryCleaner dirCleaner = null;
			if (nThreads * nRendersPerThread > 1000) {
				final long threeMins = 3 * 60 * 1000;
				
				System.out.println("A cleaner thread is being launched to clean up files greater " +
						"than 3 minutes old since this kind of test is assumed to be a stamina test " +
						"rather than a disk filler test.");
				dirCleaner = new DirectoryCleaner(root, threeMins);
				Thread t = new Thread(dirCleaner, "directoryCleaner");
				t.setDaemon(true);
				t.start();
			}
			
			final ConversionInstruction instruction = new ConversionInstruction();
			instruction.setConversionFormats(new ConversionFormat[]{
					ConversionFormat.FORMAT_ODT,
					ConversionFormat.FORMAT_WORD,
					ConversionFormat.FORMAT_PDF,
			});
			instruction.setOutputFileName("exampleTemplate");
			instruction.setCleanupDataProvider(false);

	    	for(int i =0 ; i< nThreads; i++) {
	    		final int idx = i;
	    		final String threadName = "parallelRender" + i;
		    	threads[idx] = new Thread(threadName) {
		    		public void run() {
		    			try {
		    				for(int j=0;j<nRendersPerThread;j++){
		    					renderDoc(new File(root, threadName+"."+j + ".zip"), outputStats, dp, instruction);
								log.info("render completed.  Iteration " + j);
		    				}
						} catch (Exception e) {
							throw new RuntimeException(e);
						}
		    		}
		    	};
	    	}

	    	for (int i=0;i < threads.length; i++) {
	    		threads[i].start();
	    	}
	    	
	    	for (int i=0;i < threads.length; i++) {
	    		try {
					threads[i].join();
				} catch (InterruptedException e) {
					// so be it
				}
	    	}
	    	
	    	if (dirCleaner != null) {
	        	dirCleaner.shutdown();
	    		System.out.println("cleaner cleaned up files:" + dirCleaner.getFilesRemoved());
	    		System.out.println("cleaner failed to clean up files:" + dirCleaner.getFilesUnableToRemove());
	    	}
	    	
	    	long millis = System.currentTimeMillis() - start;
	    	
	    	int nRenders = nThreads * nRendersPerThread;
	    	double nDocsPerSec = (nRenders/(millis/1000.0));
	    	
	    	log.info("Rendered " + nRenders + " in " + millis + " millis ("+ nDocsPerSec + " per sec)");
		} finally {
			if (statsWriter != null) {
				statsWriter.close();
			}
		}
    	
	}
	
	/**
	 * Build a DataProvider.  There's several ways to do this.
	 * 
	 * @return a new data provider that will provide data for the template.
	 */
	private static DataProvider buildDataProvider()
	{
		DataProviderBuilder fda = new DataProviderBuilder();
		
		// Use the DataProviderBuilder to build the data provider
		// from a String array.
		fda.addAll(new String[][] {
				{"projectName", "Super Demo Project"},
				{"name", " Freddy James"},
				{"addr", " my home pad"},
				{"DOB", " 10 July 1990"},
				{"name.1", " Paul Stuo"},
				{"addr.1", " 10 the lane"},
				{"DOB.1", " 10 Jan 1956"},
				{"name.2", " Sam Wells"},
				{"addr.2", ""},
				{"DOB.2", " 1 Apr 2000"},
				{"name.3", " Superman"},
				{"addr.3", " Level 28, The Daily Planet"},
				{"DOB.3", " 23 June 1964"},
				{"name.4", " Freddy Kruger"},
				{"addr.4", " 10 Alexander Heights"},
				{"DOB.4", " 28 May 2001"},
				{"list.funThings.0", " touch rugby"},
				{"list.funThings.1", " golf"},
				{"list.funThings.2", " ice cream"},
				{"show1", "[[true]]"},
				{"repeat1.messages.0", "This is the first of a few messages."},
				{"repeat1.messages.1", "Here's a second just for the fun of it."},
				{"repeat1.messages.2", "Phew, this is getting boring."},
				{"repeat1.messages.3", "Right that's enough."},
		});
		
		// Could also do it this way but it's a bit more manual.
//		MemoryDataProvider mdp = new MemoryDataProvider();
//		mdp.addString("projectName", "Super Demo Project");
//		MemoryDataProvider sub = new MemoryDataProvider();
//		sub.addString("messages", "This is the first of a few messages.");
//		sub.addString("messages", "Here's a second just for the fun of it.");
//		sub.addString("messages", "Phew, this is getting boring.");
//		sub.addString("messages", "Right that's enough.");
//		mdp.addDataProvider("repeat1", sub);
		
		return fda.getDataProvider();
	}
	
	/**
	 * @param args
	 * @throws ProcessingException 
	 * @throws TemplateStoreException 
	 * @throws ConverterPoolException 
	 * @throws ConversionException 
	 * @throws IOException
	 * @throws NoConvertersRunningException
	 */
	public static void main(String[] args)
	throws IOException, ConversionException, ConverterPoolException, TemplateStoreException,
	ProcessingException, URISyntaxException, NoConvertersRunningException
	{
		if (args == null || args.length > 3 || args.length == 1 ||
				(args.length == 3 && !"-s".equals(args[0]))) {
			printUsageAndExit();
			System.exit(1);
		}

		// initialise the system based on configuration
    	SystemManager.initialise();

    	// load the data from the sampleData.txt file
    	DataProvider dp = null;
		URL dataUrl = ClassLoader.getSystemResource("sampleData.txt");
		if (dataUrl == null) {
			log.info("Unable to locate sampleData.txt in the classpath, creating data in code");
			dp = buildDataProvider();
		} else {
			File dataFile = new File(new URI(dataUrl.toString()).getPath());
			dp = new DataProviderBuilder().addFile(dataFile, '|').getDataProvider();
		}
		
		// Add a system-wide renderer known as "alt1" that alternates colours for rows 
		DocumentProcessor.setDefaultRenderer("alt1", new FieldRenderer() {

			// anonymous subclass to change the default behaviour of just alternating the 
			// row colour.
			public RenderedField render(FieldDetails details, RenderedField field)
			{
				if ("value".equals(details.getFieldName()) && "Beta".equals(details.getFieldValue())) {
					field.setBold(true);
					field.setItalic(true);
					field.setUnderline(true);
					field.setCellBackgroundColour("#ffdddd");
					field.setValue("Beta (with rendering changes)");
				}
				
				return field;
			}
		});


		try {
			// load any templates in the current directory into the template store
			prepareTemplates();
			
			if (args.length == 0) {
				// run a single document render
				runExampleRender(dp);
			} else {
				try {
					// run a multithreaded multidocument render
					int index = 0;
					boolean outputStats = false;
					if (args.length == 3) {
						outputStats = true;
						index++;
					}
					int nThreads = Integer.parseInt(args[index++]);
					int nDocsPerThread = Integer.parseInt(args[index++]);
					
					if (nThreads * nDocsPerThread > 1000000) {
						System.out.println("This test case is limited to 1,000,000 documents total.");
						printUsageAndExit();
					}
					
					runConcurrentRenders(nThreads, nDocsPerThread, outputStats, dp);
					
				} catch(NumberFormatException e) {
					System.out.println("Unknown number encountered.");
					printUsageAndExit();
				}
			}
		} finally {

			// shutdown the system
			SystemManager.release();
			if (dp instanceof Cleanable) {
				((Cleanable)dp).cleanup();
			}
		}
	}

	private static void printUsageAndExit()
	{
		System.err.println("usage: java " + ExampleRenders.class.getName() + " [-h] [nThreads nDocsPerThread]");
		System.err.println("  -h        : print this usage");
		System.err.println("  <no args> : run a simple test render (creates newDocuments.zip)");
		System.err.println("  n m       : run a load test of n threads rendering m documents per thread (creates renders directory)");
		System.err.println("            : (you may want to change your converterPool.xml configuration for this test).");
		System.err.println("  -s n m    : Run the same load test with statistics");
		System.err.println("Each document render takes the exampleTemplate.doc template and the sampleData.txt file and creates ");
		System.err.println("a zip file containing the resulting document in PDF, DOC and ODF formats.");
		System.exit(1);
	}
	
	private static class DirectoryCleaner implements Runnable
	{
		private static final long SLEEP_DURATION = 60000;
		
		private File dir;
		private long maxAge;
		private int filesRemoved;
		private int filesUnableToRemove;
		private boolean running;
		
		public DirectoryCleaner(File dir, long maxAge)
		{
			this.dir = dir;
			this.maxAge = maxAge;
		}
		
		public int getFilesRemoved() {
			return filesRemoved;
		}
		
		public int getFilesUnableToRemove()
		{
			return filesUnableToRemove;
		}
		
		public void shutdown()
		{
			running = false;
			
			// give the 60 seconds sleep grace to complete.
        	try{Thread.sleep(SLEEP_DURATION);}catch(InterruptedException e) {}
		}
		
		public void run() 
		{
			running = true;
			try {
				while(running) {
					final long maxAgeTime = System.currentTimeMillis() - maxAge;
					
					dir.listFiles(new FileFilter() {
						public boolean accept(File pathname)
						{
							if (pathname.lastModified() < maxAgeTime) {
								if (pathname.delete()) {
									filesRemoved++;
								} else {
									filesUnableToRemove++;
								}
							}
							return false;
						}
						
					});
	
					Thread.sleep(SLEEP_DURATION);
				}
			
			} catch(InterruptedException e) {
				//
			}
		}
	}
}
